//! Lock Token module
use crate::TxId;
use borsh::{BorshDeserialize, BorshSerialize};
use chrono::{DateTime, Utc};
use hmac::{Hmac, Mac};
use serde::{Deserialize, Serialize};
use sha2::Sha256;
use std::{ops::Deref, time::Duration};

type HmacSha256 = Hmac<Sha256>;

#[derive(thiserror::Error, Debug, Serialize)]
/// Error type
pub enum Error {
    /// The token is expired
    #[error("Expired token")]
    Expired,

    /// Invalid signature
    #[error("Invalid signature")]
    InvalidSignature,

    /// I/O error
    #[error("IO error: {0}")]
    #[serde(serialize_with = "crate::serialize_error_to_string")]
    IO(#[from] std::io::Error),
}

#[derive(Debug)]
/// Lock Token Manager
pub struct TokenManager(pub Vec<u8>);

impl TokenManager {
    /// Creates a new instance `Token` from a vector of bytes,
    ///
    /// The token is verified to be valid and to be still not expired before it is being returned.
    pub fn verify(&self, transaction_id: &TxId, token: &[u8]) -> Result<Token, Error> {
        let token: Token = borsh::BorshDeserialize::try_from_slice(token)?;

        let mut mac = HmacSha256::new_from_slice(&self.0).expect("HMAC can take key of any size");
        mac.update(&token.inner.to_bytes().unwrap_or_default());
        let result = mac.finalize().into_bytes();
        if &result[..] != token.proof || token.inner.transaction != *transaction_id {
            Err(Error::InvalidSignature)
        } else {
            token.is_valid().map(|_| token)
        }
    }

    /// Creates a new instance of the token
    pub fn new_token(&self, transaction: &TxId, duration: Duration) -> Token {
        let token = InnerToken {
            transaction: transaction.clone(),
            owner: "owner".to_string(),
            expires_at: Utc::now() + duration,
        };
        let token_bytes = token.to_bytes().unwrap_or_default();
        let mut mac = HmacSha256::new_from_slice(&self.0).expect("HMAC can take key of any size");
        mac.update(&token_bytes);
        let proof = mac.finalize().into_bytes();

        Token {
            inner: token,
            proof: proof.to_vec(),
        }
    }
}

#[derive(Serialize, Deserialize, BorshSerialize, BorshDeserialize, Debug, PartialEq)]
/// Token to lock a transaction for a period of time
pub struct Token {
    #[serde(flatten)]
    inner: InnerToken,
    proof: Vec<u8>,
}

impl Deref for Token {
    type Target = InnerToken;

    fn deref(&self) -> &Self::Target {
        &self.inner
    }
}

#[derive(Serialize, Deserialize, BorshSerialize, BorshDeserialize, Debug, PartialEq)]
/// Inner token
///
/// This token has the transaction ID, the expiration date and the owner string. This is the data to
/// be signed by the HMAC and sent to the client.
pub struct InnerToken {
    transaction: TxId,
    #[borsh(
        serialize_with = "crate::to_ts_microseconds",
        deserialize_with = "crate::from_ts_microseconds"
    )]
    expires_at: DateTime<Utc>,
    owner: String,
}

impl InnerToken {
    /// Converts the token to bytes
    pub fn to_bytes(&self) -> Result<Vec<u8>, Error> {
        Ok(borsh::to_vec(&self)?)
    }

    /// Checks if the token is valid
    pub fn is_valid(&self) -> Result<(), Error> {
        if self.expires_at < Utc::now() {
            Err(Error::Expired)
        } else {
            Ok(())
        }
    }
}
